#ifndef _BASIC_H
#define _BASIC_H

#include <stdint.h>

#define TOKEN_EOL             0
#define TOKEN_IDENT           1	// special case - identifier follows
#define TOKEN_INTEGER	        2	// special case - integer follows (line numbers only)
#define TOKEN_NUMBER	        3	// special case - number follows
#define TOKEN_STRING	        4	// special case - string follows

#define TOKEN_LBRACKET        8
#define TOKEN_RBRACKET        9
#define TOKEN_PLUS            10
#define TOKEN_MINUS           11
#define TOKEN_MULT            12
#define TOKEN_DIV             13
#define TOKEN_EQUALS	        14
#define TOKEN_GT              15
#define TOKEN_LT              16
#define TOKEN_NOT_EQ          17
#define TOKEN_GT_EQ           18
#define TOKEN_LT_EQ           19
#define TOKEN_CMD_SEP         20
#define TOKEN_SEMICOLON       21
#define TOKEN_COMMA           22
#define TOKEN_AND             23	// FIRST_IDENT_TOKEN
#define TOKEN_OR              24
#define TOKEN_NOT             25
#define TOKEN_PRINT           26
#define TOKEN_LET             27
#define TOKEN_LIST            28
#define TOKEN_RUN             29
#define TOKEN_GOTO            30
#define TOKEN_REM             31
#define TOKEN_STOP            32
#define TOKEN_INPUT           33
#define TOKEN_CONT            34
#define TOKEN_IF              35
#define TOKEN_THEN            36
#define TOKEN_LEN             37
#define TOKEN_VAL             38
#define TOKEN_RND             39
#define TOKEN_INT             40
#define TOKEN_STR             41
#define TOKEN_FOR             42
#define TOKEN_TO              43
#define TOKEN_STEP            44
#define TOKEN_NEXT            45
#define TOKEN_MOD             46
#define TOKEN_NEW             47
#define TOKEN_GOSUB           48
#define TOKEN_RETURN          49
#define TOKEN_DIM             50
#define TOKEN_LEFT            51
#define TOKEN_RIGHT           52
#define TOKEN_MID             53
#define TOKEN_CLS             54
#define TOKEN_PAUSE           55
#define TOKEN_POSITION        56
#define TOKEN_PIN             57
#define TOKEN_PINMODE         58
#define TOKEN_INKEY           59
#define TOKEN_SAVE            60
#define TOKEN_LOAD            61
#define TOKEN_PINREAD         62
#define TOKEN_ANALOGRD        63
#define TOKEN_DIR             64
#define TOKEN_DELETE          65
#define TOKEN_SIN             66
#define TOKEN_COS             67
#define TOKEN_ABS             68
#define TOKEN_LN              69
#define TOKEN_EXP             70
#define TOKEN_SQR             71
#define TOKEN_SGN             72
#define TOKEN_BYE             73
#define TOKEN_MSAVE           74
#define TOKEN_MLOAD           75
#define TOKEN_BATT            76
#define TOKEN_ATN             77

#define TOKEN_Q               78              // simple print from console, ex: Q 1+2  mrb 20230818
#define TOKEN_REBOOT          79              // 20230822 mrb
#define TOKEN_STATUS          80              // 20230826 mrb
#define TOKEN_TAN             81              // 20230828 mrb
#define TOKEN_SLEEP           82              // 20230908 mrb alias for BYE
#define TOKEN_DEGREES         83              // 20230913 mrb manage trig functions
#define TOKEN_RADIANS         84              // 20230914 mrb

#define FIRST_IDENT_TOKEN                 23
#define LAST_IDENT_TOKEN                  84  // must match the main .ino TOKEN_BUF_SIZE

#define FIRST_NON_ALPHA_TOKEN             8
#define LAST_NON_ALPHA_TOKEN              22

#define ERROR_NONE                        0
// parse errors
#define ERROR_LEXER_BAD_NUM               1
#define ERROR_LEXER_TOO_LONG              2
#define ERROR_LEXER_UNEXPECTED_INPUT      3
#define ERROR_LEXER_UNTERMINATED_STRING   4
#define ERROR_EXPR_MISSING_BRACKET        5
#define ERROR_UNEXPECTED_TOKEN            6
#define ERROR_EXPR_EXPECTED_NUM           7
#define ERROR_EXPR_EXPECTED_STR           8
#define ERROR_LINE_NUM_TOO_BIG            9
// runtime errors
#define ERROR_OUT_OF_MEMORY               10
#define ERROR_EXPR_DIV_ZERO               11
#define ERROR_VARIABLE_NOT_FOUND          12
#define ERROR_UNEXPECTED_CMD              13
#define ERROR_BAD_LINE_NUM                14
#define ERROR_BREAK_PRESSED               15
#define ERROR_NEXT_WITHOUT_FOR            16
#define ERROR_STOP_STATEMENT              17
#define ERROR_MISSING_THEN                18
#define ERROR_RETURN_WITHOUT_GOSUB        19
#define ERROR_WRONG_ARRAY_DIMENSIONS      20
#define ERROR_ARRAY_SUBSCRIPT_OUT_RANGE   21
#define ERROR_STR_SUBSCRIPT_OUT_RANGE     22
#define ERROR_IN_VAL_INPUT                23
#define ERROR_BAD_PARAMETER               24
#define ERROR_MSAVE                       25
#define ERROR_MLOAD                       26

#define MAX_IDENT_LEN                     8
#define MAX_NUMBER_LEN                    10

#define MEMORY_SIZE                       8192

// value (int) returned from parseXXXXX
#define ERROR_MASK						0x0FFF
#define TYPE_MASK             0xF000
#define TYPE_NUMBER						0x0000
#define TYPE_STRING						0x1000

#define IS_TYPE_NUM(x) ((x & TYPE_MASK) == TYPE_NUMBER)
#define IS_TYPE_STR(x) ((x & TYPE_MASK) == TYPE_STRING)

extern unsigned char mem[];
extern int sysPROGEND;
extern int sysSTACKSTART;
extern int sysSTACKEND;
extern int sysVARSTART;
extern int sysVAREND;
extern int sysGOSUBSTART;
extern int sysGOSUBEND;
extern uint16_t lineNumber;                 // 0 = input buffer

// forward declarations
int parseBatt(void);
void softwareReset();
int systemStatus();
void printTokens(unsigned char *);
void listProg(uint16_t first, uint16_t last);
unsigned char *findProgLine(uint16_t targetLineNumber);
void deleteProgLine(unsigned char *p);
int doProgLine(uint16_t lineNumber, unsigned char* tokenPtr, int tokensLength);
int stackPushNum(float val);
float stackPopNum();
int stackPushStr(char *str);
char *stackGetStr();
char *stackPopStr();
void stackAdd2Strs();
void stackLeftOrRightStr(int len, int mode);
void stackMidStr(int start, int len);
unsigned char *findVariable(char *searchName, int searchMask);
void deleteVariableAt(unsigned char *pos);
int storeNumVariable(char *name, float val);
int storeForNextVariable(char *name, float start, float step, float end, uint16_t lineNum, uint16_t stmtNum);
int storeStrVariable(char *name, char *val);
int createArray(char *name, int isString);
int _getArrayElemOffset(unsigned char **p, int *pOffset);
int setNumArrayElem(char *name, float val);
int setStrArrayElem(char *name);
float lookupNumArrayElem(char *name, int *error);
float lookupNumVariable(char *name);
char *lookupStrVariable(char *name);
int gosubStackPush(int lineNumber, int stmtNumber);
int gosubStackPop(int *lineNumber, int *stmtNumber);
int nextToken();
int tokenize(unsigned char *input, unsigned char *output, int outputSize);
int getNextToken();
int parseNumberExpr();
int parseSubscriptExpr();
int parseFnCallExpr();
int parseStringExpr();
int parseParenExpr();
int parse_RND();
uint16_t getBatStat(void);
int parse_INKEY();
int parseUnaryNumExp();
int parsePrimary();
int getTokPrecedence();
int parseBinOpRHS(int ExprPrec, int lhsVal);
int parseExpression();
int expectNumber();
int parse_RUN();
int parse_GOTO();
int parse_PAUSE();
int parse_LIST();
int parse_PRINT();
int parseTwoIntCmd();
int parseAssignment(bool inputStmt);
int parse_IF();
int parse_FOR();
int parse_NEXT();
int parse_GOSUB();
int parseLoadSaveCmd();
int parseMsave();
int parseMload();
int parseSimpleCmd();
int parse_DIM();
int parseStmts();
int processInput(unsigned char *tokenBuf);
void reset();
static char executeMode;
int parseExpression();
int parsePrimary();
int expectNumber();

typedef struct {
  float val;
  float step;
  float end;
  uint16_t lineNumber;
  uint16_t stmtNumber;
}
ForNextData;

typedef struct {
  char *token;
  uint8_t format;
}
TokenTableEntry;

extern const char* const errorTable[];

void reset();
int tokenize(unsigned char *input, unsigned char *output, int outputSize);
int processInput(unsigned char *tokenBuf);

#endif
